//
//  RTcmixPlayer.m
//  "base" class -- use this for iRTcmix audio work
//	will function in both AudioQueueBuffer and AudioUnit mode
//
//  Created by Brad Garton on 7/17/09.
//	modified with RTcmixManager by Damon Holzborn 3/2010
//	additional work done by Brad Garton, 4/2010
//



#include <AudioToolbox/AudioToolbox.h>
#import "RTcmixPlayer.h"
#import "RTcmixManager.h"

@implementation RTcmixPlayer

@synthesize bufferByteSize;
@synthesize numPacketsPerCallback;
@synthesize gain;
@synthesize audioPlayerShouldStopImmediately;
@synthesize queueObject;
@synthesize audioFormat;
// AudioUnit stuff
@synthesize inputProc;
@synthesize rioUnit;
@synthesize outbufdex;
@synthesize rtcmixoutbuf;
@synthesize inbufdex;
@synthesize rtcmixinbuf;
@synthesize audiounitflag;
@synthesize inputflag;
// soundfile and buffer playback stuff
@synthesize soundFilePath;



#pragma mark playback and interruption callbacks

// This is the callback function that does the sample-crunching work
// the AudioQueueBuffers get filled and sent off from here
static void playbackCallback (
	void					*inUserData,
	AudioQueueRef			inAudioQueue,
	AudioQueueBufferRef		bufferReference
) {
	// This callback, being 'outside' the implementation, needs references to the
	// RTcmixPlayer and RTcmixManager objects
	RTcmixPlayer *player = (RTcmixPlayer *) inUserData;
	RTcmixManager *rtcmixManager = [RTcmixManager sharedManager];
		
	UInt32 numBytes = [player bufferByteSize];
	UInt32 numPackets = [player numPacketsPerCallback];

	bufferReference->mAudioDataByteSize			= numBytes;		
	bufferReference->mPacketDescriptionCount	= numPackets;

	// BGG -- this drives the RTcmix sample-computation buffer
	short *bptr = bufferReference->mAudioData; // fill the AudioQueueBuffer directly in RTcmix
	pullTraverse(NULL, bptr); // no input audio, use the AudioUnit version
	
	if (check_bang() == 1) { // check for retriggering, set by "MAXBANG" in the score
		[rtcmixManager checkBangMethod];
	}

	AudioQueueEnqueueBuffer (
		inAudioQueue,
		bufferReference,
		0,
		NULL
	);
}

// this callback is used by RemoteIO AudioUnit if AudioUnit is employed instead of AudioQueueBuffers
static OSStatus	rtcmixPerformCallback (
	void						*inRefCon, 
	AudioUnitRenderActionFlags 	*ioActionFlags, 
	const AudioTimeStamp 		*inTimeStamp, 
	UInt32 						inBusNumber, 
	UInt32 						inNumberFrames, 
	AudioBufferList 			*ioData)
{
	RTcmixPlayer *player = (RTcmixPlayer *) inRefCon;
	RTcmixManager *rtcmixManager = [RTcmixManager sharedManager];

	AudioUnitRender([player rioUnit], ioActionFlags, inTimeStamp, 1, inNumberFrames, ioData);

	// these buffers are used because RTcmix is working asynch with the AudioUnit
	short *rtcmixoutptr = player.rtcmixoutbuf; // pointers to the bufs we use
	short *rtcmixinptr = NULL;
	if (player.inputflag == 1) rtcmixinptr = player.rtcmixinbuf;

	//loop through all the buffers that need to be filled
	for (int i = 0 ; i < ioData->mNumberBuffers; i++){
		//get the buffer to be filled
		AudioBuffer buffer = ioData->mBuffers[i];
		UInt16 *frameBuffer = buffer.mData;

		//loop through the buffer and fill the frames
		for (int j = 0; j < inNumberFrames*NCHANS; j += NCHANS){
			for (int k = 0; k < NCHANS; k++) {
				if(player.outbufdex >= NUMPACKETSPERCALLBACK*NCHANS) {
					pullTraverse(rtcmixinptr, rtcmixoutptr);
					if (check_bang() == 1) { // check for retriggering, set by "MAXBANG" in the score
						[rtcmixManager checkBangMethod];
					}
					player.outbufdex = 0;
					player.inbufdex = 0;
				}
				if (player.inputflag == 1)
					rtcmixinptr[player.inbufdex+k] = frameBuffer[j+k];
				frameBuffer[j+k] = rtcmixoutptr[player.outbufdex+k];
			}
			player.outbufdex += NCHANS;
			player.inbufdex += NCHANS;
		}
	}

	// used to return an OSStatus err from AudioUnitRender but it wouldna go laddie!
    return noErr;
}

void interruptionListenerCallback (
	void	*inUserData,
	UInt32	interruptionState
) {
	RTcmixPlayer *player = (RTcmixPlayer *) inUserData;	

	if (interruptionState == kAudioSessionBeginInterruption) {
		[player pause];
	} else if (interruptionState == kAudioSessionEndInterruption) {
		[player resume];
	}
}

void propListenerCallback(
	void *					inClientData,
	AudioSessionPropertyID	inID,
	UInt32                  inDataSize,
	const void *            inData)
{
	RTcmixPlayer *player = (RTcmixPlayer *) inClientData;
	if (inID == kAudioSessionProperty_AudioRouteChange)
	{
		// if there was a route change, we need to dispose the current rio unit and create a new one
		AudioComponentInstanceDispose([player rioUnit]);		

		[player setPlaybackStuff];
		[player setupRemoteIO];

		AudioOutputUnitStart([player rioUnit]);
	}
}


#pragma mark -

#pragma mark RTcmix data-setting methods

// use this to get data in via the makeconnection() PField scheme
- (void) setInlet:(int)inlet withVal:(Float32)val {
	pfield_set(inlet, val);	
}

- (int) setSoundFile: (NSString *)sfname {

	soundFilePath = [[NSBundle mainBundle]pathForResource:sfname ofType:NULL];

	if (soundFilePath == nil) {
		NSLog(@"problem setting soundfile path in mainBundle");
		return -1;
	}

	return 1;
}

- (int) loadSampleBuffer: (NSString *)bufname {
	AudioFileID     soundFile;
	AudioStreamBasicDescription		dataFormat;
	OSStatus status;
	UInt32 propsize;

	NSURL *soundFileURL = [NSURL fileURLWithPath:soundFilePath];

	// BGG -- should probably use ExtAudioFileOpenURL()
	status = AudioFileOpenURL((CFURLRef)soundFileURL, kAudioFileReadPermission, kAudioFileAIFFType, &soundFile);
	if (status != 0) {
		NSLog(@"error opening soundfile");
		return -1;
	}

	propsize = sizeof(dataFormat);
	status = AudioFileGetProperty(soundFile, kAudioFilePropertyDataFormat,
		&propsize, &dataFormat);
	if (status != 0) {
		NSLog(@"problem geting AudioFile properties");
		return -1;
	}
	//NSLog(@"SR: %f  chans: %d", dataFormat.mSampleRate, dataFormat.mChannelsPerFrame);

	//Get file size
	UInt64 outDataSize = 0;
	propsize = sizeof(UInt64);
	AudioFileGetProperty(soundFile, kAudioFilePropertyAudioDataByteCount,
		&propsize, &outDataSize);
	//NSLog(@"file size: %d", outDataSize);

	//Get packet count
	UInt64 outPackets = 0;
	propsize = sizeof(UInt64);
	AudioFileGetProperty(soundFile, kAudioFilePropertyAudioDataPacketCount,
		&propsize, &outPackets);
	//	NSLog(@"file packets: %d", outPackets);

	UInt32 numBytes = outDataSize;
	UInt32 numPackets = outPackets;
	sampBuf = malloc(outDataSize);
	if (sampBuf == NULL) {
		NSLog(@"error mallocing sample buffer");
		return -1;
	}

	status = AudioFileReadPackets (soundFile, false, &numBytes, NULL, 0, &numPackets, sampBuf);
	if (status !=0) {
		NSLog(@"problem reading file packets into buffer");
		return -1;
	}
	//	NSLog(@"numPackets read into sampBuf: %d", numPackets);

	fsampBuf = malloc(numPackets*dataFormat.mChannelsPerFrame*sizeof(float));
	if (fsampBuf == NULL) {
		NSLog(@"error mallocing fsample buffer");
		return -1;
	}

	// convert to floating-point buffer for RTcmix
	float *fsbptr = fsampBuf;
	short *sbptr = sampBuf;
	for(int i = 0; i < (numPackets*dataFormat.mChannelsPerFrame); i++) {
		sbptr[i] = CFSwapInt16(sbptr[i]);
		fsbptr[i] = (float)sbptr[i]/32768.0;
	}

	buffer_set((char *)[bufname UTF8String], fsbptr, numPackets, dataFormat.mChannelsPerFrame, 0);

	return 1;
}


#pragma mark -

#pragma mark Setup

- (id) initPlayer {

	self = [super init];

	if (self != nil) {
		// initialize RTcmix when the app gets inited
		rtcmixmain();
		
		// before instantiating the playback audio queue object, 
		//	set the audio session category
		UInt32 sessionCategory = kAudioSessionCategory_MediaPlayback;
		AudioSessionSetProperty (kAudioSessionProperty_AudioCategory,
								 sizeof (sessionCategory), &sessionCategory);		
		
		// initialize the audio session object for this application,
		//		registering the callback that Audio Session Services will invoke 
		//		when there's an interruption
		AudioSessionInitialize (NULL, NULL, interruptionListenerCallback, self);

		audiounitflag = 0; // not running as AU

		[self setPlaybackStuff];
		[self setupPlaybackAudioQueueObject];
		[self setAudioPlayerShouldStopImmediately: NO];
		AudioSessionSetActive (true);
		[self play];
	}

	return self;
}

- (id) initPlayerAUWithInputFlag:(int)iflag {

	audiounitflag = 1; // running as AU
	inputflag = iflag; // 1 if we are running input

	self = [super init];

	if (self != nil) {
		inputProc.inputProc = rtcmixPerformCallback;
		inputProc.inputProcRefCon = self;

		// initialize the audio session object for this application,
		//		registering the callback that Audio Session Services will invoke 
		//		when there's an interruption
		AudioSessionInitialize(NULL, NULL, interruptionListenerCallback, self);
		AudioSessionSetActive(true);

		//	set the audio session category
		UInt32 sessionCategory;
		if (inputflag == 1)
			sessionCategory = kAudioSessionCategory_PlayAndRecord;
		else
			sessionCategory = kAudioSessionCategory_MediaPlayback;

		AudioSessionSetProperty(kAudioSessionProperty_AudioCategory,
			sizeof(sessionCategory), &sessionCategory);


		if (audiounitflag == 1) { //possibly do these for AudioQueue version too?
			// add a property listener
			AudioSessionAddPropertyListener(kAudioSessionProperty_AudioRouteChange,
				propListenerCallback, self);
	
			// this sets buffer size when using AudioUnit
			Float32 preferredBufferSize = .1;
			AudioSessionSetProperty(kAudioSessionProperty_PreferredHardwareIOBufferDuration,
				sizeof(preferredBufferSize), &preferredBufferSize);
		}

		Float64 prefSR = 44100.0;
		AudioSessionSetProperty(kAudioSessionProperty_PreferredHardwareSampleRate,
			sizeof(prefSR), &prefSR);

		// this is because input AudioUnits default to the phone/speaker output, and I want to hear it!
		UInt32 newRoute = kAudioSessionOverrideAudioRoute_Speaker;
		AudioSessionSetProperty(kAudioSessionProperty_OverrideAudioRoute,
			sizeof(newRoute), &newRoute);


		[self setPlaybackStuff];
		[self setupRemoteIO];


		// initialize RTcmix when the app gets inited
		rtcmixmain();
		// set rtcmix params (note we don't use inbuf and outbuf, they are passed in pulltraverse() above)
		maxmsp_rtsetparams(audioFormat.mSampleRate, audioFormat.mChannelsPerFrame, NUMPACKETSPERCALLBACK, NULL, NULL, errbuf);

		outbufdex = NUMPACKETSPERCALLBACK*NCHANS; // set at max to get a pullTraverse() straightaway

		// set up the buffers used by RTcmix (asynch with RemoteIO
		rtcmixoutbuf = malloc(NUMPACKETSPERCALLBACK*NCHANS*sizeof(short)); 
		if (rtcmixoutbuf == NULL) // ok, not much of a check...
			NSLog(@"rtcmixoutbuf is NULL!");

		inbufdex = NUMPACKETSPERCALLBACK*NCHANS;
		if (inputflag == 1) {
			rtcmixinbuf = malloc(NUMPACKETSPERCALLBACK*NCHANS*sizeof(short));
			if (rtcmixinbuf == NULL)
				NSLog(@"rtcmixinbuf is NULL!");

			// zero out input buffer
			memset((void *)rtcmixinbuf, 0, NUMPACKETSPERCALLBACK*NCHANS*sizeof(short));
		}

		[self setGain: 1.0];
		[self setAudioPlayerShouldStopImmediately: NO];
		[self play];
	}

	return self;
}

- (void) setPlaybackStuff {

	memset((void *)&audioFormat, 0, sizeof(audioFormat)); // recommended

	Float64 hardwareSampleRate;
	UInt32 propertySize = sizeof(hardwareSampleRate);	
	AudioSessionGetProperty ( kAudioSessionProperty_CurrentHardwareSampleRate,
		&propertySize, &hardwareSampleRate);

	if (inputflag == 1)
		audioFormat.mSampleRate = hardwareSampleRate; // we have to use hardware rate for input
	else
		audioFormat.mSampleRate = 44100.0;

	audioFormat.mFormatID = kAudioFormatLinearPCM;
	audioFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked;
	audioFormat.mBitsPerChannel = 16;
	audioFormat.mChannelsPerFrame = NCHANS;
	audioFormat.mBytesPerFrame = NCHANS*2; // assumes 16-bit samples, 2 bytes
	audioFormat.mFramesPerPacket = 1;
	audioFormat.mBytesPerPacket = audioFormat.mBytesPerFrame * audioFormat.mFramesPerPacket;
	audioFormat.mReserved = 0;
}

- (int) setupRemoteIO
{	
	// Open the output unit
	AudioComponentDescription desc;
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_RemoteIO;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;

	AudioComponent comp = AudioComponentFindNext(NULL, &desc);
	AudioComponentInstanceNew(comp, &rioUnit);

#define kOutputBus 0
#define kInputBus 1

	UInt32 one = 1;
	AudioUnitSetProperty(rioUnit, kAudioOutputUnitProperty_EnableIO,
		kAudioUnitScope_Output, kOutputBus, &one, sizeof(one));
	AudioUnitSetProperty(rioUnit, kAudioUnitProperty_StreamFormat,
		kAudioUnitScope_Input, kOutputBus, &audioFormat, sizeof(audioFormat));
	AudioUnitSetProperty(rioUnit, kAudioUnitProperty_SetRenderCallback,
		kAudioUnitScope_Global, kOutputBus, &inputProc, sizeof(inputProc));

	if (inputflag == 1) // enable input side
		AudioUnitSetProperty(rioUnit, kAudioOutputUnitProperty_EnableIO,
			kAudioUnitScope_Input, kInputBus, &one, sizeof(one));

	AudioUnitInitialize(rioUnit);

	return 0;
}

- (void) setupPlaybackAudioQueueObject {

	// create the playback audio queue object
	AudioQueueNewOutput (
		&audioFormat,
		playbackCallback,
		self, 
		CFRunLoopGetCurrent (),
		kCFRunLoopCommonModes,
		0,								// run loop flags
		&queueObject
	);
	
	// set the volume of the playback audio queue
	[self setGain: 1.0];
	
	AudioQueueSetParameter (
		queueObject,
		kAudioQueueParam_Volume,
		gain
	);
}

- (void) setupAudioQueueBuffers {

	// set basic queue/callback params
	[self setNumPacketsPerCallback: NUMPACKETSPERCALLBACK];
	[self setBufferByteSize: [self numPacketsPerCallback] * audioFormat.mBytesPerFrame];

	// set rtcmix params (note we don't use inbuf and outbuf, they are passed in pulltraverse() above)
	maxmsp_rtsetparams(audioFormat.mSampleRate, audioFormat.mChannelsPerFrame, NUMPACKETSPERCALLBACK, NULL, NULL, errbuf);

	// prime the queue with some data before starting
	// allocate and enqueue buffers				
	int bufferIndex;
	for (bufferIndex = 0; bufferIndex < kNumberAudioDataBuffers; ++bufferIndex) {
		AudioQueueAllocateBuffer (
			[self queueObject],
			[self bufferByteSize],
			&buffers[bufferIndex]
		);

		playbackCallback ( 
			self,
			[self queueObject],
			buffers[bufferIndex]
		);
	}
}

- (void) newGain:(Float32)newgain {
	// set the volume of the playback audio queue
	[self setGain: newgain];
	
	AudioQueueSetParameter (
		queueObject,
		kAudioQueueParam_Volume,
		gain
	);
}

- (void) play {

	if (audiounitflag == 0) {
		[self setupAudioQueueBuffers];

		AudioQueueStart (
			self.queueObject,
			NULL			// start time. NULL means ASAP.
		);
	} else {
		AudioOutputUnitStart([self rioUnit]);
	}
}

- (void) stop {

	if (audiounitflag == 0) {
		AudioQueueStop (
			self.queueObject,
			self.audioPlayerShouldStopImmediately
		);
	} else {
		AudioOutputUnitStop([self rioUnit]);
	}
}


- (void) pause {

	if (audiounitflag == 0) {
		AudioQueuePause (
			self.queueObject
		);
	} else {
		AudioOutputUnitStop([self rioUnit]);
	}
}


- (void) resume {

	UInt32 sessionCategory;
	sessionCategory = kAudioSessionCategory_MediaPlayback;
	if (inputflag == 1) 
		sessionCategory = kAudioSessionCategory_PlayAndRecord;

	AudioSessionSetProperty (
							 kAudioSessionProperty_AudioCategory,
							 sizeof (sessionCategory),
							 &sessionCategory
							 );
	AudioSessionSetActive (true);

	if (audiounitflag == 0) {
		AudioQueueStart (
			self.queueObject,
			NULL			// start time. NULL means ASAP
		);
	} else {
		AudioOutputUnitStart([self rioUnit]);
	}
}


#pragma mark -

- (void) dealloc {

	AudioQueueDispose (
		queueObject, 
		YES
	);
	
	[super dealloc];
}

@end
